import io
import os
import shutil
import stat

import six
from crontab import CronTab

from dagster import DagsterInstance, check, utils
from dagster.core.host_representation import ExternalSchedule
from dagster.core.scheduler import DagsterSchedulerError, Scheduler
from dagster.serdes import ConfigurableClass


class SystemCronScheduler(Scheduler, ConfigurableClass):
    '''Scheduler implementation that uses the local systems cron. Only works on unix systems that
    have cron.

    Enable this scheduler by adding it to your ``dagster.yaml`` in ``$DAGSTER_HOME``.
    '''

    def __init__(
        self, inst_data=None,
    ):
        self._inst_data = inst_data

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return SystemCronScheduler(inst_data=inst_data)

    def get_cron_tab(self):
        return CronTab(user=True)

    def debug_info(self):
        return "Running Cron Jobs:\n{jobs}\n".format(
            jobs="\n".join(
                [str(job) for job in self.get_cron_tab() if 'dagster-schedule:' in job.comment]
            )
        )

    def start_schedule(self, instance, external_schedule):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.inst_param(external_schedule, 'external_schedule', ExternalSchedule)
        schedule_origin_id = external_schedule.get_origin_id()

        # If the cron job already exists, remove it. This prevents duplicate entries.
        # Then, add a new cron job to the cron tab.
        if self.running_schedule_count(external_schedule.get_origin_id()) > 0:
            self._end_cron_job(instance, schedule_origin_id)

        self._start_cron_job(instance, external_schedule)

        # Verify that the cron job is running
        running_schedule_count = self.running_schedule_count(schedule_origin_id)
        if running_schedule_count == 0:
            raise DagsterSchedulerError(
                "Attempted to write cron job for schedule "
                "{schedule_name}, but failed. "
                "The scheduler is not running {schedule_name}.".format(
                    schedule_name=external_schedule.name
                )
            )
        elif running_schedule_count > 1:
            raise DagsterSchedulerError(
                "Attempted to write cron job for schedule "
                "{schedule_name}, but duplicate cron jobs were found. "
                "There are {running_schedule_count} jobs running for the schedule."
                "To resolve, run `dagster schedule up`, or edit the cron tab to "
                "remove duplicate schedules".format(
                    schedule_name=external_schedule.name,
                    running_schedule_count=running_schedule_count,
                )
            )

    def stop_schedule(self, instance, schedule_origin_id):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.str_param(schedule_origin_id, 'schedule_origin_id')

        schedule = self._get_schedule_state(instance, schedule_origin_id)

        self._end_cron_job(instance, schedule_origin_id)

        # Verify that the cron job has been removed
        running_schedule_count = self.running_schedule_count(schedule_origin_id)
        if running_schedule_count > 0:
            raise DagsterSchedulerError(
                "Attempted to remove existing cron job for schedule "
                "{schedule_name}, but failed. "
                "There are still {running_schedule_count} jobs running for the schedule.".format(
                    schedule_name=schedule.name, running_schedule_count=running_schedule_count
                )
            )

    def wipe(self, instance):
        # Note: This method deletes schedules from ALL repositories
        check.inst_param(instance, 'instance', DagsterInstance)

        # Delete all script files
        script_directory = os.path.join(instance.schedules_directory(), "scripts")
        if os.path.isdir(script_directory):
            shutil.rmtree(script_directory)

        # Delete all logs
        logs_directory = os.path.join(instance.schedules_directory(), "logs")
        if os.path.isdir(logs_directory):
            shutil.rmtree(logs_directory)

        # Remove all cron jobs
        with self.get_cron_tab() as cron_tab:
            for job in cron_tab:
                if 'dagster-schedule:' in job.comment:
                    cron_tab.remove_all(comment=job.comment)

    def _get_bash_script_file_path(self, instance, schedule_origin_id):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.str_param(schedule_origin_id, 'schedule_origin_id')

        script_directory = os.path.join(instance.schedules_directory(), "scripts")
        utils.mkdir_p(script_directory)

        script_file_name = "{}.sh".format(schedule_origin_id)
        return os.path.join(script_directory, script_file_name)

    def _cron_tag_for_schedule(self, schedule_origin_id):
        return 'dagster-schedule: {schedule_origin_id}'.format(
            schedule_origin_id=schedule_origin_id
        )

    def _get_command(self, script_file, instance, schedule_origin_id):
        schedule_log_file_path = self.get_logs_path(instance, schedule_origin_id)
        command = "{script_file} > {schedule_log_file_path} 2>&1".format(
            script_file=script_file, schedule_log_file_path=schedule_log_file_path
        )

        return command

    def _start_cron_job(self, instance, external_schedule):
        schedule_origin_id = external_schedule.get_origin_id()
        script_file = self._write_bash_script_to_file(instance, external_schedule)
        command = self._get_command(script_file, instance, schedule_origin_id)

        with self.get_cron_tab() as cron_tab:
            job = cron_tab.new(
                command=command,
                comment='dagster-schedule: {schedule_origin_id}'.format(
                    schedule_origin_id=schedule_origin_id
                ),
            )
            job.setall(external_schedule.cron_schedule)

    def _end_cron_job(self, instance, schedule_origin_id):
        with self.get_cron_tab() as cron_tab:
            cron_tab.remove_all(comment=self._cron_tag_for_schedule(schedule_origin_id))

        script_file = self._get_bash_script_file_path(instance, schedule_origin_id)
        if os.path.isfile(script_file):
            os.remove(script_file)

    def running_schedule_count(self, schedule_origin_id):
        matching_jobs = self.get_cron_tab().find_comment(
            self._cron_tag_for_schedule(schedule_origin_id)
        )

        return len(list(matching_jobs))

    def _get_or_create_logs_directory(self, instance, schedule_origin_id):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.str_param(schedule_origin_id, 'schedule_origin_id')

        logs_directory = os.path.join(instance.schedules_directory(), "logs", schedule_origin_id)
        if not os.path.isdir(logs_directory):
            utils.mkdir_p(logs_directory)

        return logs_directory

    def get_logs_path(self, instance, schedule_origin_id):
        check.inst_param(instance, 'instance', DagsterInstance)
        check.str_param(schedule_origin_id, 'schedule_origin_id')

        logs_directory = self._get_or_create_logs_directory(instance, schedule_origin_id)
        return os.path.join(logs_directory, 'scheduler.log')

    def _write_bash_script_to_file(self, instance, external_schedule):
        # Get path to store bash script
        schedule_origin_id = external_schedule.get_origin_id()
        script_file = self._get_bash_script_file_path(instance, schedule_origin_id)

        # Get path to store schedule attempt logs
        logs_directory = self._get_or_create_logs_directory(instance, schedule_origin_id)
        schedule_log_file_name = "{}_{}.result".format("${RUN_DATE}", schedule_origin_id)
        schedule_log_file_path = os.path.join(logs_directory, schedule_log_file_name)

        local_target = external_schedule.get_origin()

        # Environment information needed for execution
        dagster_home = os.getenv('DAGSTER_HOME')

        script_contents = '''
            #!/bin/bash
            export DAGSTER_HOME={dagster_home}
            export LANG=en_US.UTF-8
            {env_vars}

            export RUN_DATE=$(date "+%Y%m%dT%H%M%S")

            {python_exe} -m dagster api launch_scheduled_execution --schedule_name {schedule_name} {repo_cli_args} "{result_file}"
        '''.format(
            python_exe=local_target.executable_path,
            schedule_name=external_schedule.name,
            repo_cli_args=local_target.get_repo_cli_args(),
            result_file=schedule_log_file_path,
            dagster_home=dagster_home,
            env_vars="\n".join(
                [
                    "export {key}={value}".format(key=key, value=value)
                    for key, value in external_schedule.environment_vars.items()
                ]
            ),
        )

        with io.open(script_file, 'w', encoding='utf-8') as f:
            f.write(six.text_type(script_contents))

        st = os.stat(script_file)
        os.chmod(script_file, st.st_mode | stat.S_IEXEC)

        return script_file



import os
import re
import subprocess

import pytest
import yaml
from dagster_cron import SystemCronScheduler

from dagster import ScheduleDefinition
from dagster.core.code_pointer import FileCodePointer
from dagster.core.definitions import lambda_solid, pipeline, repository
from dagster.core.host_representation import PythonEnvRepositoryLocation, RepositoryLocationHandle
from dagster.core.instance import DagsterInstance, InstanceType
from dagster.core.launcher.sync_in_memory_run_launcher import SyncInMemoryRunLauncher
from dagster.core.scheduler import ScheduleStatus
from dagster.core.scheduler.scheduler import (
    DagsterScheduleDoesNotExist,
    DagsterScheduleReconciliationError,
    DagsterSchedulerError,
)
from dagster.core.storage.event_log import InMemoryEventLogStorage
from dagster.core.storage.noop_compute_log_manager import NoOpComputeLogManager
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import InMemoryRunStorage
from dagster.core.storage.schedules import SqliteScheduleStorage
from dagster.core.test_utils import environ
from dagster.seven import TemporaryDirectory


@pytest.fixture(scope='function')
def restore_cron_tab():
    with TemporaryDirectory() as tempdir:
        crontab_backup = os.path.join(tempdir, "crontab_backup.txt")
        with open(crontab_backup, 'wb+') as f:
            try:
                output = subprocess.check_output(['crontab', '-l'])
                f.write(output)
            except subprocess.CalledProcessError:
                # If a crontab hasn't been created yet, the command fails with a
                # non-zero error code
                pass

        try:
            subprocess.check_output(['crontab', '-r'])
        except subprocess.CalledProcessError:
            # If a crontab hasn't been created yet, the command fails with a
            # non-zero error code
            pass

        yield

        subprocess.check_output(['crontab', crontab_backup])


@pytest.fixture(scope='function')
def unset_dagster_home():
    old_env = os.getenv('DAGSTER_HOME')
    if old_env is not None:
        del os.environ['DAGSTER_HOME']
    yield
    if old_env is not None:
        os.environ['DAGSTER_HOME'] = old_env


@pipeline
def no_config_pipeline():
    @lambda_solid
    def return_hello():
        return 'Hello'

    return return_hello()


schedules_dict = {
    'no_config_pipeline_daily_schedule': ScheduleDefinition(
        name="no_config_pipeline_daily_schedule",
        cron_schedule="0 0 * * *",
        pipeline_name="no_config_pipeline",
        run_config={"storage": {"filesystem": None}},
    ),
    'no_config_pipeline_every_min_schedule': ScheduleDefinition(
        name="no_config_pipeline_every_min_schedule",
        cron_schedule="* * * * *",
        pipeline_name="no_config_pipeline",
        run_config={"storage": {"filesystem": None}},
    ),
    'default_config_pipeline_every_min_schedule': ScheduleDefinition(
        name="default_config_pipeline_every_min_schedule",
        cron_schedule="* * * * *",
        pipeline_name="no_config_pipeline",
    ),
}


def define_schedules():
    return list(schedules_dict.values())


@repository
def test_repository():
    if os.getenv('DAGSTER_TEST_SMALL_REPO'):
        return [no_config_pipeline] + list(
            filter(
                lambda x: not x.name == "default_config_pipeline_every_min_schedule",
                define_schedules(),
            )
        )

    return [no_config_pipeline] + define_schedules()


def get_test_external_repo():
    return PythonEnvRepositoryLocation(
        RepositoryLocationHandle.create_out_of_process_location(
            location_name='test_location',
            repository_code_pointer_dict={
                'test_repository': FileCodePointer(__file__, 'test_repository'),
            },
        )
    ).get_repository('test_repository')


def get_smaller_external_repo():
    with environ({'DAGSTER_TEST_SMALL_REPO': '1'}):
        return get_test_external_repo()


def get_cron_jobs():
    output = subprocess.check_output(['crontab', '-l'])
    return list(filter(None, output.decode('utf-8').strip().split("\n")))


def define_scheduler_instance(tempdir):
    return DagsterInstance(
        instance_type=InstanceType.EPHEMERAL,
        local_artifact_storage=LocalArtifactStorage(tempdir),
        run_storage=InMemoryRunStorage(),
        event_storage=InMemoryEventLogStorage(),
        compute_log_manager=NoOpComputeLogManager(),
        schedule_storage=SqliteScheduleStorage.from_local(os.path.join(tempdir, 'schedules')),
        scheduler=SystemCronScheduler(),
        run_launcher=SyncInMemoryRunLauncher(),
    )


def test_init(restore_cron_tab):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repository = get_test_external_repo()
        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repository)

        # Check schedules are saved to disk
        assert 'schedules' in os.listdir(tempdir)

        assert instance.all_stored_schedule_state()


def test_re_init(restore_cron_tab):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        # Start schedule
        schedule_state = instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        # Re-initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        # Check schedules are saved to disk
        assert 'schedules' in os.listdir(tempdir)

        schedule_states = instance.all_stored_schedule_state()

        for state in schedule_states:
            if state.name == "no_config_pipeline_every_min_schedule":
                assert state == schedule_state


def test_start_and_stop_schedule(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        schedule = external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        schedule_origin_id = schedule.get_origin_id()

        instance.start_schedule_and_update_storage_state(schedule)

        assert 'schedules' in os.listdir(tempdir)

        assert "{}.sh".format(schedule_origin_id) in os.listdir(
            os.path.join(tempdir, 'schedules', 'scripts')
        )

        instance.stop_schedule_and_update_storage_state(schedule_origin_id)

        assert "{}.sh".format(schedule_origin_id) not in os.listdir(
            os.path.join(tempdir, 'schedules', 'scripts')
        )


def test_start_non_existent_schedule(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)

        with pytest.raises(DagsterScheduleDoesNotExist):
            # Initialize scheduler
            instance.stop_schedule_and_update_storage_state("asdf")


def test_start_schedule_cron_job(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule")
        )
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("default_config_pipeline_every_min_schedule")
        )

        # Inspect the cron tab
        cron_jobs = get_cron_jobs()

        assert len(cron_jobs) == 3

        external_schedules_dict = {
            external_repo.get_external_schedule(name).get_origin_id(): schedule_def
            for name, schedule_def in schedules_dict.items()
        }

        for cron_job in cron_jobs:
            match = re.findall(r"^(.*?) (/.*) > (.*) 2>&1 # dagster-schedule: (.*)", cron_job)
            cron_schedule, command, log_file, schedule_origin_id = match[0]

            schedule_def = external_schedules_dict[schedule_origin_id]

            # Check cron schedule matches
            if schedule_def.cron_schedule == "0 0 * * *":
                assert cron_schedule == "@daily"
            else:
                assert cron_schedule == schedule_def.cron_schedule

            # Check bash file exists
            assert os.path.isfile(command)

            # Check log file is correct
            assert log_file.endswith("scheduler.log")


def test_remove_schedule_def(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        assert len(instance.all_stored_schedule_state()) == 3

        instance.reconcile_scheduler_state(get_smaller_external_repo())

        assert len(instance.all_stored_schedule_state()) == 2


def test_add_schedule_def(restore_cron_tab):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_smaller_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(get_smaller_external_repo())

        # Start all schedule and verify cron tab, schedule storage, and errors
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule")
        )
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        assert len(instance.all_stored_schedule_state()) == 2
        assert len(get_cron_jobs()) == 2
        assert len(instance.scheduler_debug_info().errors) == 0

        # Reconcile with an additional schedule added
        instance.reconcile_scheduler_state(get_test_external_repo())

        assert len(instance.all_stored_schedule_state()) == 3
        assert len(get_cron_jobs()) == 2
        assert len(instance.scheduler_debug_info().errors) == 0

        # Start the new schedule
        external_repo = get_test_external_repo()
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("default_config_pipeline_every_min_schedule")
        )

        assert len(instance.all_stored_schedule_state()) == 3
        assert len(get_cron_jobs()) == 3
        assert len(instance.scheduler_debug_info().errors) == 0


def test_start_and_stop_schedule_cron_tab(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()
        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        # Start schedule
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 1

        # Try starting it again
        with pytest.raises(DagsterSchedulerError):
            instance.start_schedule_and_update_storage_state(
                external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
            )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 1

        # Start another schedule
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule")
        )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 2

        # Stop second schedule
        instance.stop_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule").get_origin_id()
        )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 1

        # Try stopping second schedule again
        instance.stop_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule").get_origin_id()
        )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 1

        # Start second schedule
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule")
        )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 2

        # Reconcile schedule state, should be in the same state
        instance.reconcile_scheduler_state(external_repo)
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 2

        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("default_config_pipeline_every_min_schedule")
        )
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 3

        # Reconcile schedule state, should be in the same state
        instance.reconcile_scheduler_state(external_repo)
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 3

        # Stop all schedules
        instance.stop_schedule_and_update_storage_state(
            external_repo.get_external_schedule(
                "no_config_pipeline_every_min_schedule"
            ).get_origin_id()
        )
        instance.stop_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_daily_schedule").get_origin_id()
        )
        instance.stop_schedule_and_update_storage_state(
            external_repo.get_external_schedule(
                "default_config_pipeline_every_min_schedule"
            ).get_origin_id()
        )

        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 0

        # Reconcile schedule state, should be in the same state
        instance.reconcile_scheduler_state(external_repo)
        cron_jobs = get_cron_jobs()
        assert len(cron_jobs) == 0


def test_script_execution(
    restore_cron_tab, unset_dagster_home
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        os.environ["DAGSTER_HOME"] = tempdir
        config = {
            'scheduler': {'module': 'dagster_cron', 'class': 'SystemCronScheduler', 'config': {}},
            # This needs to synchronously execute to completion when
            # the generated bash script is invoked
            'run_launcher': {
                'module': 'dagster.core.launcher.sync_in_memory_run_launcher',
                'class': 'SyncInMemoryRunLauncher',
            },
        }

        with open(os.path.join(tempdir, 'dagster.yaml'), 'w+') as f:
            f.write(yaml.dump(config))

        instance = DagsterInstance.get()
        external_repo = get_test_external_repo()
        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        schedule_origin_id = external_repo.get_external_schedule(
            "no_config_pipeline_every_min_schedule"
        ).get_origin_id()
        script = instance.scheduler._get_bash_script_file_path(  # pylint: disable=protected-access
            instance, schedule_origin_id
        )

        subprocess.check_output([script], shell=True, env={"DAGSTER_HOME": tempdir})

        runs = instance.get_runs()
        assert len(runs) == 1
        assert runs[0].status == PipelineRunStatus.SUCCESS


def test_start_schedule_fails(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        def raises(*args, **kwargs):
            raise Exception('Patch')

        instance._scheduler._start_cron_job = raises  # pylint: disable=protected-access
        with pytest.raises(Exception, match='Patch'):
            instance.start_schedule_and_update_storage_state(
                external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
            )

        schedule = instance.get_schedule_state(
            external_repo.get_external_schedule(
                "no_config_pipeline_every_min_schedule"
            ).get_origin_id()
        )

        assert schedule.status == ScheduleStatus.STOPPED


def test_start_schedule_unsuccessful(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        def do_nothing(*_):
            pass

        instance._scheduler._start_cron_job = do_nothing  # pylint: disable=protected-access

        # End schedule
        with pytest.raises(
            DagsterSchedulerError,
            match="Attempted to write cron job for schedule no_config_pipeline_every_min_schedule, "
            "but failed. The scheduler is not running no_config_pipeline_every_min_schedule.",
        ):
            instance.start_schedule_and_update_storage_state(
                external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
            )


def test_start_schedule_manual_delete_debug(
    restore_cron_tab, snapshot  # pylint:disable=unused-argument,redefined-outer-name
):
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        # Manually delete the schedule from the crontab
        instance.scheduler._end_cron_job(  # pylint: disable=protected-access
            instance,
            external_repo.get_external_schedule(
                "no_config_pipeline_every_min_schedule"
            ).get_origin_id(),
        )

        # Check debug command
        debug_info = instance.scheduler_debug_info()
        assert len(debug_info.errors) == 1

        # Reconcile should fix error
        instance.reconcile_scheduler_state(external_repo)
        debug_info = instance.scheduler_debug_info()
        assert len(debug_info.errors) == 0


def test_start_schedule_manual_add_debug(
    restore_cron_tab, snapshot  # pylint:disable=unused-argument,redefined-outer-name
):
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        # Manually add the schedule from to the crontab
        instance.scheduler._start_cron_job(  # pylint: disable=protected-access
            instance, external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        # Check debug command
        debug_info = instance.scheduler_debug_info()
        assert len(debug_info.errors) == 1

        # Reconcile should fix error
        instance.reconcile_scheduler_state(external_repo)
        debug_info = instance.scheduler_debug_info()
        assert len(debug_info.errors) == 0


def test_start_schedule_manual_duplicate_schedules_add_debug(
    restore_cron_tab, snapshot  # pylint:disable=unused-argument,redefined-outer-name
):
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)
        external_schedule = external_repo.get_external_schedule(
            "no_config_pipeline_every_min_schedule"
        )

        instance.start_schedule_and_update_storage_state(external_schedule)

        # Manually add  extra cron tabs
        instance.scheduler._start_cron_job(  # pylint: disable=protected-access
            instance, external_schedule,
        )
        instance.scheduler._start_cron_job(  # pylint: disable=protected-access
            instance, external_schedule,
        )

        # Check debug command
        debug_info = instance.scheduler_debug_info()
        assert len(debug_info.errors) == 1

        # Reconcile should fix error
        instance.reconcile_scheduler_state(external_repo)
        debug_info = instance.scheduler_debug_info()
        assert len(debug_info.errors) == 0


def test_stop_schedule_fails(
    restore_cron_tab,  # pylint:disable=unused-argument,redefined-outer-name
):
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        external_schedule = external_repo.get_external_schedule(
            "no_config_pipeline_every_min_schedule"
        )
        schedule_origin_id = external_schedule.get_origin_id()

        def raises(*args, **kwargs):
            raise Exception('Patch')

        instance._scheduler._end_cron_job = raises  # pylint: disable=protected-access

        instance.start_schedule_and_update_storage_state(external_schedule)

        assert 'schedules' in os.listdir(tempdir)

        assert "{}.sh".format(schedule_origin_id) in os.listdir(
            os.path.join(tempdir, 'schedules', 'scripts')
        )

        # End schedule
        with pytest.raises(Exception, match='Patch'):
            instance.stop_schedule_and_update_storage_state(schedule_origin_id)

        schedule = instance.get_schedule_state(schedule_origin_id)

        assert schedule.status == ScheduleStatus.RUNNING


def test_stop_schedule_unsuccessful(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        def do_nothing(*_):
            pass

        instance._scheduler._end_cron_job = do_nothing  # pylint: disable=protected-access

        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        # End schedule
        with pytest.raises(
            DagsterSchedulerError,
            match="Attempted to remove existing cron job for schedule "
            "no_config_pipeline_every_min_schedule, but failed. There are still 1 jobs running for "
            "the schedule.",
        ):
            instance.stop_schedule_and_update_storage_state(
                external_repo.get_external_schedule(
                    "no_config_pipeline_every_min_schedule"
                ).get_origin_id()
            )


def test_wipe(restore_cron_tab):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        # Start schedule
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        # Wipe scheduler
        instance.wipe_all_schedules()

        # Check schedules are wiped
        assert instance.all_stored_schedule_state() == []


def test_log_directory(restore_cron_tab):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:

        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()
        external_schedule = external_repo.get_external_schedule(
            "no_config_pipeline_every_min_schedule"
        )
        schedule_log_path = instance.logs_path_for_schedule(external_schedule.get_origin_id())

        assert schedule_log_path.endswith(
            "/schedules/logs/{schedule_origin_id}/scheduler.log".format(
                schedule_origin_id=external_schedule.get_origin_id()
            )
        )

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        # Start schedule
        instance.start_schedule_and_update_storage_state(external_schedule)

        # Wipe scheduler
        instance.wipe_all_schedules()

        # Check schedules are wiped
        assert instance.all_stored_schedule_state() == []


def test_reconcile_failure(restore_cron_tab):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        instance.reconcile_scheduler_state(external_repo)
        instance.start_schedule_and_update_storage_state(
            external_repo.get_external_schedule("no_config_pipeline_every_min_schedule")
        )

        def failed_start_job(*_):
            raise DagsterSchedulerError("Failed to start")

        def failed_end_job(*_):
            raise DagsterSchedulerError("Failed to stop")

        instance._scheduler.start_schedule = failed_start_job  # pylint: disable=protected-access
        instance._scheduler.stop_schedule = failed_end_job  # pylint: disable=protected-access

        # Initialize scheduler
        with pytest.raises(
            DagsterScheduleReconciliationError,
            match="Error 1: Failed to stop\n    Error 2: Failed to stop\n    Error 3: Failed to stop",
        ):
            instance.reconcile_scheduler_state(external_repo)


def test_reconcile_failure_when_deleting_schedule_def(
    restore_cron_tab,
):  # pylint:disable=unused-argument,redefined-outer-name
    with TemporaryDirectory() as tempdir:
        instance = define_scheduler_instance(tempdir)
        external_repo = get_test_external_repo()

        # Initialize scheduler
        instance.reconcile_scheduler_state(external_repo)

        assert len(instance.all_stored_schedule_state()) == 3

        def failed_end_job(*_):
            raise DagsterSchedulerError("Failed to stop")

        instance._scheduler.stop_schedule_and_delete_from_storage = (  # pylint: disable=protected-access
            failed_end_job
        )

        with pytest.raises(
            DagsterScheduleReconciliationError, match="Error 1: Failed to stop",
        ):
            instance.reconcile_scheduler_state(get_smaller_external_repo())
